<?php

require_once 'Attribute.php';
require_once 'Record.php';
require_once 'RecordSet.php';
require_once 'Wikidata.php';

/**
 * @file
 *
 * Manage internal transactions (NOT mysql transactions... confuzzeled yet?)
 *
 * To use:
 *
 * startNewTransaction($userId, $userIP, $comment, $dc)
 * then do a getUpdateTransactionId() to find the id you need for
 * add_transaction_id.
 *
 * Since this is not a mysql transaction, I don't THINK you need
 * to close it. If you do, it wasn't documented.
 *
 * There's also something to do with restrictions. This is the only
 * documentation you get so far. You're on your own now :-P
 * (Please document anything else that's really needed for anything)
 */

interface QueryTransactionInformation {
	public function getRestriction( Table $table );

	public function getTables();

	public function versioningAttributes();

	public function versioningFields( $tableName );

	public function versioningOrderBy();

	public function versioningGroupBy( Table $table );

	public function setVersioningAttributes( Record $record, $row );
}

class DefaultQueryTransactionInformation implements QueryTransactionInformation {
	public function getRestriction( Table $table ) {
		return "1";
	}

	public function getTables() {
		return [];
	}

	public function versioningAttributes() {
		return [];
	}

	public function versioningFields( $tableName ) {
		return [];
	}

	public function versioningOrderBy() {
		return [];
	}

	public function versioningGroupBy( Table $table ) {
		return [];
	}

	public function setVersioningAttributes( Record $record, $row ) {
	}

	public function __toString() {
		return "QueryTransactionInformation (...)";
	}
}

class QueryLatestTransactionInformation extends DefaultQueryTransactionInformation {
	public function getRestriction( Table $table ) {
		return getLatestTransactionRestriction( $table->getIdentifier() );
	}

	public function setVersioningAttributes( Record $record, $row ) {
	}
}

class QueryHistoryTransactionInformation extends DefaultQueryTransactionInformation {
	public function versioningAttributes() {
		$o = OmegaWikiAttributes::getInstance();

		return [ $o->recordLifeSpan ];
	}

	public function versioningFields( $tableName ) {
		return [ $tableName . '.add_transaction_id', $tableName . '.remove_transaction_id', $tableName . '.remove_transaction_id IS NULL AS is_live' ];
	}

	public function versioningOrderBy() {
		return [ 'is_live DESC', 'add_transaction_id DESC' ];
	}

	public function setVersioningAttributes( Record $record, $row ) {
		$o = OmegaWikiAttributes::getInstance();

		$record->recordLifeSpan = getRecordLifeSpanTuple( $row['add_transaction_id'], $row['remove_transaction_id'] );
	}
}

class QueryAtTransactionInformation extends DefaultQueryTransactionInformation {
	protected $transactionId;
	protected $addAttributes;

	public function __construct( $transactionId, $addAttributes ) {
		$this->transactionId = $transactionId;
		$this->addAttributes = $addAttributes;
	}

	public function getRestriction( Table $table ) {
		return getAtTransactionRestriction( $table->getIdentifier(), $this->transactionId );
	}

	public function versioningAttributes() {
		$o = OmegaWikiAttributes::getInstance();

		if ( $this->addAttributes ) {
			return [ $o->recordLifeSpan ];
		} else {
			return [];
		}
	}

	public function versioningFields( $tableName ) {
		return [ $tableName . '.add_transaction_id', $tableName . '.remove_transaction_id', $tableName . '.remove_transaction_id IS NULL AS is_live' ];
	}

	public function setVersioningAttributes( Record $record, $row ) {
		$o = OmegaWikiAttributes::getInstance();

		if ( $this->addAttributes ) {
			$record->recordLifeSpan = getRecordLifeSpanTuple( $row['add_transaction_id'], $row['remove_transaction_id'] );
		}
	}
}

class QueryUpdateTransactionInformation extends DefaultQueryTransactionInformation {
	protected $transactionId;

	public function __construct( $transactionId ) {
		$this->transactionId = $transactionId;
	}

	public function getRestriction( Table $table ) {
		return " " . $table->getIdentifier() . ".add_transaction_id =" . $this->transactionId .
			" OR " . $table->getIdentifier() . ".removeTransactionId =" . $this->transactionId;
	}

	// public function versioningAttributes() {
	// global
	// $recordLifeSpanAttribute;
	//
	// return array();
	// }

	// public function versioningFields($tableName) {
	// return array($tableName . '.add_transaction_id', $tableName . '.remove_transaction_id', $tableName . '.remove_transaction_id IS NULL AS is_live');
	// }

	// public function setVersioningAttributes($record, $row) {
	// global
	// $recordLifeSpanAttribute;
	//
	// $record->setAttributeValue($recordLifeSpanAttribute, getRecordLifeSpanTuple($row['add_transaction_id'], $row['remove_transaction_id']));
	// }
}

global
	$updateTransactionId;

function startNewTransaction( $userID, $userIP, $comment, $dc = null ) {
	global
		$updateTransactionId;

	if ( $dc === null ) {
		$dc = wdGetDataSetContext();
	}

	$dbw = wfGetDB( DB_PRIMARY );
	$timestamp = wfTimestampNow();

	// do not store IP for logged in users
	if ( $userID > 0 ) {
		$userIP = "";
	}

	$dbw->insert(
		"{$dc}_transactions",
		[ 'user_id' => $userID,
			'user_ip' => $userIP,
			'timestamp' => $timestamp,
			'comment' => $comment
		], __METHOD__
	);
	$updateTransactionId = $dbw->insertId();
	// return is not really needed, as the global variable is set,
	// but this solves the case where $trans = startNewTransaction(...) would be used
	return $updateTransactionId;
}

function getUpdateTransactionId() {
	global $updateTransactionId, $wgRequest;

	if ( !isset( $updateTransactionId ) ) {
		// normally startNewTransaction is invoqued before this function
		// but actually we can call it from here directly, it should work the same
		global $wgUser;
		startNewTransaction( $wgUser->getID(), $wgRequest->getIP(), '' );
	}

	return $updateTransactionId;
}

function getLatestTransactionId() {
	$dc = wdGetDataSetContext();
	$dbr = wfGetDB( DB_REPLICA );
	// false if not found
	$transactionId = $dbr->selectField(
		"{$dc}_transactions",
		'MAX(transaction_id)',
		'',
		__METHOD__
	);

	if ( $transactionId ) {
		return $transactionId;
	}
	return 0;
}

function getLatestTransactionRestriction( $table ) {
	return ' ' . $table . '.remove_transaction_id IS NULL ';
}

function getAtTransactionRestriction( $table, $transactionId ) {
	return ' ' . $table . '.add_transaction_id <= ' . $transactionId . ' AND (' .
				$table . '.remove_transaction_id > ' . $transactionId . ' OR ' . $table . '.remove_transaction_id IS NULL) ';
}

function getViewTransactionRestriction( $table ) {
	global
		$wgRequest;

	$action = $wgRequest->getText( 'action' );

	if ( $action == 'edit' ) {
		return getLatestTransactionRestriction( $table );
	} elseif ( $action == 'history' ) {
		return '1';
	} else {
		return getLatestTransactionRestriction( $table );
	}
}

function getOperationSelectColumn( $table, $transactionId ) {
	return " IF($table.add_transaction_id=$transactionId, 'Added', 'Removed') AS operation ";
}

function getInTransactionRestriction( $table, $transactionId ) {
	return " ($table.add_transaction_id=$transactionId OR $table.remove_transaction_id=$transactionId) ";
}

function getUserName( $userId ) {
	$dbr = wfGetDB( DB_REPLICA );
	$userName = $dbr->selectField(
		'user',
		'user_name',
		[
			'user_id' => $userId
		], __METHOD__
	);

	if ( $userName ) {
		return $userName;
	}
	return '';
}

function getUserLabel( $userId, $userIP ) {
	if ( $userId > 0 ) {
		return getUserName( $userId );
	} elseif ( $userIP != "" ) {
		return $userIP;
	} else {
		return "Unknown";
	}
}

function expandUserIDsInRecordSet( RecordSet $recordSet, Attribute $userID, Attribute $userIP ) {
	for ( $i = 0; $i < $recordSet->getRecordCount(); $i++ ) {
		$record = $recordSet->getRecord( $i );
		$record->setAttributeValue(
			$userIDAttribute,
			getUserLabel(
				$record->$userIDAttribute,
				$record->$userIP
			)
		);
	}
}

function expandTransactionIdsInRecordSet( RecordSet $recordSet ) {
	for ( $i = 0; $i < $recordSet->getRecordCount(); $i++ ) {
		$record = $recordSet->getRecord( $i );
		$record->transaction = getTransactionRecord( $record->transactionId );
	}
}

function getTransactionRecord( $transactionId ) {
	$o = OmegaWikiAttributes::getInstance();

	$dc = wdGetDataSetContext();
	$result = new ArrayRecord( $o->transactionStructure );
	$result->transactionId = $transactionId;

	if ( $transactionId > 0 ) {
		$transaction = OwDatabaseAPI::getTransactionIdDetails( $transactionId, [], $dc );

		if ( $transaction ) {
			$result->user = getUserLabel( $transaction->user_id, $transaction->user_ip );
			if ( $result->user == null ) { $result->user = "userId " . $transaction->user_id . " not found";
			}
			$result->timestamp = $transaction->timestamp;
			$result->summary = $transaction->comment;
		}
	} else {
		if ( $transactionId != null ) {
			$result->user = "Unknown";
		} else {
			$result->user = "";
		}

		$result->timestamp = "";
		$result->summary = "";
	}

	return $result;
}

function getRecordLifeSpanTuple( $addTransactionId, $removeTransactionId ) {
	$o = OmegaWikiAttributes::getInstance();

	$result = new ArrayRecord( $o->recordLifeSpanStructure );
	$result->addTransaction = getTransactionRecord( $addTransactionId );
	$result->removeTransaction = getTransactionRecord( $removeTransactionId );

	return $result;
}

function getTransactionLabel( $transactionId ) {
	$o = OmegaWikiAttributes::getInstance();

	if ( $transactionId > 0 ) {
		$record = getTransactionRecord( $transactionId );

		$label =
			timestampAsText( $record->timestamp ) . ', ' .
			$record->user;

		$summary = $record->summary;

		if ( $summary != "" ) {
			$label .= ', ' . $summary;
		}

		return $label;
	} else {
		return "";
	}
}

class Transactions {

	/**
	 * Checks the latest transaction id from syntrans (added transactions).
	 *
	 * @note Other transaction ids were excluded because of the slow speed to generate them.
	 * (any search for transaction ids which are not present will search for all
	 * the rows in a column which causes the long query time).
	 * Syntrans seems to cover most of the changes. This function will change
	 * when a new table is in place to cover this. Doing so will make the function
	 * faster (since the latest ids are already set) and accurate ( since it would
	 * not be limited to added syntrans only ).
	 *
	 * @param int $languageId
	 * @param array $options optional parameters
	 * @param string|null $dc WikiLexical dataset
	 *
	 * @return int The latest transaction_id.
	 * @return -1 If the language_id is non numeric or no transaction_id was found
	 * @return -2 if a there are any current jobs pending, this function is skipped
	 *
	 * @note options parameter can be used to extend this function.
	 * Though you can access this function, it is highly recommended that you
	 * use the static function OwDatabaseAPI::getLanguageIdLatestTransactionId instead.
	 * Also note that this function currently includes all data, even removed ones.
	 */
	public static function getLanguageIdLatestTransactionId( $languageId, $options = [], $dc = null ) {
		// If non numeric, skip this function and return -1
		if ( !is_numeric( $languageId ) ) {
			return -1;
		}

		$dbr = wfGetDB( DB_REPLICA );

		// If jobs exist, skip this function and return -2
		// unless if it is the same job that requires a transaction_id
		$jobExists = $dbr->selectField(
			"job",
			'job_id',
			null, __METHOD__,
			[
				'LIMIT' => 1
			]
		);

		if ( !isset( $options['is_the_job'] ) ) {
			$options['is_the_job'] = false;
		}

		if ( $jobExists && !$options['is_the_job'] ) {
			return -2;
		}

		if ( $dc === null ) {
			$dc = wdGetDataSetContext();
		}

		$Transaction = new Transactions;
		$transaction_id = $Transaction->getLanguageIdLatestSynonymsAndTranslationsTransactionId( $languageId );

		if ( $transaction_id ) {
			return $transaction_id;
		}
		return -1;
	}

/** @see static function Transactions::getLanguageIdLatestTransactionId
 */
	protected function getLanguageIdLatestSynonymsAndTranslationsTransactionId( $languageId, $options = [], $dc = null ) {
		if ( $dc === null ) {
			$dc = wdGetDataSetContext();
		}
		$dbr = wfGetDB( DB_REPLICA );

		$result = $dbr->selectField(
			[
				'exp' => "{$dc}_expression",
				'synt' => "{$dc}_syntrans"
			],
			'synt.add_transaction_id AS tid',
			[
				'language_id' => $languageId,
				'synt.expression_id = exp.expression_id',
				'synt.remove_transaction_id' => null,
				'exp.remove_transaction_id' => null
			], __METHOD__,
			[
				'ORDER BY' => 'syntrans_sid DESC'
			]
		);

		if ( $result ) {
			return $result;
		}
		return -1;
	}

	/**
	 * @param int $transactionId req'd The transaction id
	 * @param array $options Optional parameters
	 * @param string $dc opt'l The WikiLexicalData dataset
	 *
	 * @return array( int user_id, str user_ip, str timestamp, str comment )
	 * @return if not exists, array()
	 *
	 * @note options parameter can be used to extend this function.
	 * Though you can access this function, it is highly recommended that you
	 * use the static function OwDatabaseAPI::getTransactionIdDetails instead.
	 * Also note that this function currently includes all data, even removed ones.
	 */
	public static function getIdDetails( $transactionId, $options, $dc ) {
		if ( $dc === null ) {
			$dc = wdGetDataSetContext();
		}
		$dbr = wfGetDB( DB_REPLICA );
		$transaction = $dbr->selectRow(
			"{$dc}_transactions",
			[ 'user_id', 'user_ip', 'timestamp', 'comment' ],
			[ 'transaction_id' => $transactionId ], __METHOD__
		);

		if ( $transaction ) {
			return $transaction;
		}
		return [];
	}

}
